using System.Collections;
using System.Collections.Generic;
using UnityEngine;

/// <summary>
/// GPUInstance需要自己改写一些shader，可以看我改写的NewUnlitShader，就加几个指令就行了，有些指令不加也能用
/// </summary>
public class GPUInstance : MonoBehaviour
{
    public GameObject prefab;
    public int InstanceCount = 100;//最大1023
    
    private Mesh mesh;
    private Material material;
    private Matrix4x4[] matrix;
    private MeshFilter[] meshFilters;
    private Renderer[] renders;

    private Vector4[] colors;
    private MaterialPropertyBlock materialPropertyBlock;
    
    void Awake() {
        if(prefab == null)
            return;
        
        var meshFilter = prefab.GetComponent<MeshFilter>();
        if(meshFilter) {
            mesh = prefab.GetComponent<MeshFilter>().sharedMesh;
            material = prefab.GetComponent<Renderer>().sharedMaterial;
        }
        
        matrix = new Matrix4x4[InstanceCount];
        colors = new Vector4[InstanceCount];
        materialPropertyBlock = new MaterialPropertyBlock();
        
        for(int i = 0; i < InstanceCount; i++) {
            //随机位置
            float x = Random.Range(-50, 50);
            float y = Random.Range(-3, 3);
            float z = Random.Range(-50, 50);
            //设置旋转
            matrix[i] = Matrix4x4.identity;
            //设置位置
            matrix[i].SetColumn(3, new Vector4(x, y, z, 1));
            //设置缩放，矩阵缩放
            matrix[i].m00   = Mathf.Max(1, 1);
            matrix[i].m11   = Mathf.Max(1, 1);
            matrix[i].m22   = Mathf.Max(1, 1);
        }
    }
    
    void Update() {
        // 传入mesh、材质、矩阵
        // 可以使用 materialPropertyBlock 覆盖 material
        Graphics.DrawMeshInstanced(mesh, 0, material, matrix, matrix.Length, materialPropertyBlock);
        
        //Graphics.DrawMeshInstanced(mesh, 0, material, matrix, matrix.Length);
    }
}
