using System;
using System.Collections;
using System.Collections.Generic;
using UnityEditor;
using UnityEngine;

public class GetSkinData : MonoBehaviour
{
	//勾选这个，则将动画文件数据转为SO数据，so文件是自己提前创建好的
	//不勾选这个则是播放矩阵动画
	public bool isSetAnimSo;
	
	
	private float time = 0.0f;
    public SkinnedMeshRenderer smr;
    private GPUSkinning_Bone[] bones = null;//骨骼数据
    private Mesh mesh = null;//mesh数据
    private int rootBoneIndex = 0;//根骨骼节点
    private GPUSkinning_BoneAnimation boneAnimationsSOData = null;//动画数据



    public Transform spawnPos;
    private Material newMtrl = null;
    private Mesh newMesh = null;
    private Matrix4x4[] matricesUniformBlock = null;
    private int shaderPropID_Matrices = 0;


    public List<string> boneNameList = new List<string>();
    private int fps = 30;

    // Start is called before the first frame update
    void Start() {
	    mesh = smr.sharedMesh;
	    shaderPropID_Matrices = Shader.PropertyToID("_Matrices");
	    int numBones = smr.bones.Length;
	    matricesUniformBlock = new Matrix4x4[numBones];
	    
	    //采集数据
	    MyCode();
	    CollectionData();
	    
	    //生成SO文件
	    if (isSetAnimSo)
	    {
		    GetBoneData();
		    return;
	    }

	    //return;
	    //读取SO数据
	    GPUSkinning_BoneAnimation data = Resources.Load<GPUSkinning_BoneAnimation>("GPUSkinningData");
	    boneAnimationsSOData = data;
	    
	    
	    //材质数据
	    newMtrl = new Material(Shader.Find("Unlit/GPUSkinning"));
	    newMtrl.CopyPropertiesFromMaterial(smr.sharedMaterial);
	    
		//顶点数据
	    Vector4[] tangents = new Vector4[mesh.vertexCount];
	    for(int i = 0; i < mesh.vertexCount; ++i)
	    {
		    BoneWeight boneWeight = mesh.boneWeights[i];
		    tangents[i].x = boneWeight.boneIndex0;
		    tangents[i].y = boneWeight.weight0;
		    tangents[i].z = boneWeight.boneIndex1;
		    tangents[i].w = boneWeight.weight1;
	    }
	    
	    //mesh数据
	    newMesh = new Mesh();
	    newMesh.vertices = mesh.vertices;
	    newMesh.tangents = tangents;
	    newMesh.uv = mesh.uv;
	    newMesh.triangles = mesh.triangles;
	    
	    //生成模型
	    GPUSkinning_SpawnObject spawnObject = new GPUSkinning_SpawnObject();
	    spawnObject.transform = spawnPos;
	    spawnObject.mf = spawnObject.transform.gameObject.AddComponent<MeshFilter>();
	    spawnObject.mr = spawnObject.transform.gameObject.AddComponent<MeshRenderer>();
	    spawnObject.mr.sharedMaterial = newMtrl;
	    spawnObject.mf.sharedMesh = newMesh;
    }

    //在Update里面播放动画
    private void Update()
    {
	    if (isSetAnimSo)
	    {
		    return;
	    }
	    
	    UpdateBoneAnimationMatrix(null, time);
	    Play();
	    time += Time.deltaTime;
    }

    private void Play()
    {
	    int numBones = bones.Length;
	    for(int i = 0; i < numBones; ++i)
	    {
		    matricesUniformBlock[i] = bones[i].animationMatrix;
	    }
	    newMtrl.SetMatrixArray(shaderPropID_Matrices, matricesUniformBlock);
    }
    
    private void UpdateBoneAnimationMatrix(string animName, float time)
    {
	    //so文件里面的数据
	    //GPUSkinning_BoneAnimation boneAnimation = boneAnimationsSOData;//GetBoneAnimation(animName);
	    int frameIndex = (int)(time * fps) % (int)(boneAnimationsSOData.frames.Length);
	    GPUSkinning_BoneAnimationFrame frame = boneAnimationsSOData.frames[frameIndex];
	    
	    
	    //设置旋转问题
	    var x = Matrix4x4.TRS(transform.position, transform.rotation, transform.lossyScale);
	    //传递每帧的动画数据 和 根骨骼
	    UpdateBoneTransformMatrix(bones[rootBoneIndex], x, frame);
    }
    
    private void UpdateBoneTransformMatrix(GPUSkinning_Bone rootBone, Matrix4x4 parentMatrix, GPUSkinning_BoneAnimationFrame frame)
    {
	    int index = BoneAnimationFrameIndexOf(rootBone);

	    Matrix4x4 mat = parentMatrix * frame.matrices[index];//mat的值会有子父物体关系的依赖而产生变化
	    rootBone.animationMatrix = mat * rootBone.bindpose;

	    //寻找子骨骼应该变成的样子
	    GPUSkinning_Bone[] children = rootBone.children;
	    int numChildren = children.Length;
	    for(int i = 0; i < numChildren; ++i)
	    {
		    UpdateBoneTransformMatrix(children[i], mat, frame);
	    }
    }
    
    private int BoneAnimationFrameIndexOf(GPUSkinning_Bone bone)
    {
	    return boneNameList.IndexOf(bone.name);
    }


    private void CollectionData()
    {
	    AnimationClip animClip = GetComponent<Animator>().runtimeAnimatorController.animationClips[0];
	    EditorCurveBinding[] curvesBinding = AnimationUtility.GetCurveBindings(animClip);
	    for (int i = 0; i < curvesBinding.Length; i++)
	    {
		    var curveBinding = curvesBinding[i];
		    GPUSkinning_Bone bone = GetBoneByHierarchyName(curveBinding.path);
		    if (boneNameList.Contains(bone.name))
		    {
			    continue;
		    }
		    boneNameList.Add(bone.name);
	    }
    }
    
    
    private void GetBoneData() {
	    var data = Resources.Load<GPUSkinning_BoneAnimation>("GPUSkinningData");

	    int boneAnimationsCount = 0;
	    boneAnimationsSOData = new GPUSkinning_BoneAnimation();

	    AnimationClip animClip = GetComponent<Animator>().runtimeAnimatorController.animationClips[0];

	    //foreach()
	    {
		    GPUSkinning_BoneAnimation boneAnimation = data;
		    //设置fps
		    //boneAnimation.fps = 30;
		    //设置动画片段名
		    //boneAnimation.animName = animClip.name;
		    //设置帧数
		    boneAnimation.frames = new GPUSkinning_BoneAnimationFrame[(int)(animClip.length * fps)];
		    //boneAnimation.length = animClip.length;
		    boneAnimationsSOData = boneAnimation;

		    for (int frameIndex = 0; frameIndex < boneAnimation.frames.Length; ++frameIndex) {
			    GPUSkinning_BoneAnimationFrame frame = new GPUSkinning_BoneAnimationFrame();
			    float second = (float)(frameIndex) / (float)fps;

			    List<GPUSkinning_Bone> bones2 = new List<GPUSkinning_Bone>();
			    List<Matrix4x4> matrices = new List<Matrix4x4>();
			    List<string> bonesHierarchyNames = new List<string>();
			    EditorCurveBinding[] curvesBinding = AnimationUtility.GetCurveBindings(animClip);

			    for (int i = 0; i < curvesBinding.Length; i++)
			    {
				    var curveBinding = curvesBinding[i];
				    GPUSkinning_Bone bone = GetBoneByHierarchyName(curveBinding.path);

				    if (bones2.Contains(bone)) {
					    continue;
				    }

				    bones2.Add(bone);

				    bonesHierarchyNames.Add(GetBoneHierarchyName(bone));

				    AnimationCurve curveRX = AnimationUtility.GetEditorCurve(animClip, curveBinding.path, curveBinding.type,
					    "m_LocalRotation.x");
				    AnimationCurve curveRY = AnimationUtility.GetEditorCurve(animClip, curveBinding.path, curveBinding.type,
					    "m_LocalRotation.y");
				    AnimationCurve curveRZ = AnimationUtility.GetEditorCurve(animClip, curveBinding.path, curveBinding.type,
					    "m_LocalRotation.z");
				    AnimationCurve curveRW = AnimationUtility.GetEditorCurve(animClip, curveBinding.path, curveBinding.type,
					    "m_LocalRotation.w");

				    AnimationCurve curvePX = AnimationUtility.GetEditorCurve(animClip, curveBinding.path, curveBinding.type,
					    "m_LocalPosition.x");
				    AnimationCurve curvePY = AnimationUtility.GetEditorCurve(animClip, curveBinding.path, curveBinding.type,
					    "m_LocalPosition.y");
				    AnimationCurve curvePZ = AnimationUtility.GetEditorCurve(animClip, curveBinding.path, curveBinding.type,
					    "m_LocalPosition.z");

				    float curveRX_v = curveRX.Evaluate(second);
				    float curveRY_v = curveRY.Evaluate(second);
				    float curveRZ_v = curveRZ.Evaluate(second);
				    float curveRW_v = curveRW.Evaluate(second);

				    float curvePX_v = curvePX.Evaluate(second);
				    float curvePY_v = curvePY.Evaluate(second);
				    float curvePZ_v = curvePZ.Evaluate(second);

				    Vector3 translation = new Vector3(curvePX_v, curvePY_v, curvePZ_v);
				    Quaternion rotation = new Quaternion(curveRX_v, curveRY_v, curveRZ_v, curveRW_v);
				    NormalizeQuaternion(ref rotation);
				    matrices.Add(
					    Matrix4x4.TRS(translation, rotation, Vector3.one)
				    );
			    }




			    frame.matrices = matrices.ToArray();
			    
			    boneAnimation.frames[frameIndex] = frame;
		    }
	    }

	    //保存骨骼数据到SO文件
	    EditorUtility.SetDirty(data);
	    AssetDatabase.SaveAssets();
	    AssetDatabase.Refresh();
    }

    /// <summary>
    /// 将SkinMeshRender 的数据转换到 GPUSkinning_Bone[]，并且做好骨骼的父子关系
    /// </summary>
    private void MyCode() {
	    int numBones = smr.bones.Length;
	    bones = new GPUSkinning_Bone[numBones];
	    mesh = smr.sharedMesh;
	    for (int i = 0; i < numBones; ++i) {
		    //获取骨骼数据
		    GPUSkinning_Bone bone = new GPUSkinning_Bone();
		    bones[i] = bone;
		    //这个地方的数据又有引用
		    bone.transform = smr.bones[i];
		    bone.name = smr.bones[i].name;
		    bone.bindpose = mesh.bindposes[i] /*smr to bone*/;

		    if (smr.bones[i].transform == smr.rootBone)
		    {
			    rootBoneIndex = i;
		    }
		    
		    //Debug.Log("=====" + smr.bones[i].transform.name);
	    }
	    

	    
	    //添加骨骼的父子关系
	    System.Action<GPUSkinning_Bone> CollectChildren = null;
	    CollectChildren = (currentBone) => {
		    List<GPUSkinning_Bone> children = new List<GPUSkinning_Bone>();
		    for (int j = 0; j < currentBone.transform.childCount; ++j) {
			    Transform childTransform = currentBone.transform.GetChild(j);
			    GPUSkinning_Bone childBone = GetBoneByTransform(childTransform);
			    if (childBone != null) {
				    childBone.parent = currentBone;
				    children.Add(childBone);
				    CollectChildren(childBone);
			    }
		    }

		    currentBone.children = children.ToArray();
	    };
	    CollectChildren(bones[rootBoneIndex]);
    }

    private GPUSkinning_Bone GetBoneByHierarchyName(string hierarchyName)
    {
	    System.Func<GPUSkinning_Bone, string, GPUSkinning_Bone> Search = null;
	    Search = (bone, name) => 
	    {
		    if(name == hierarchyName)
		    {
			    return bone;
		    }
		    foreach(GPUSkinning_Bone child in bone.children)
		    {
			    GPUSkinning_Bone result = Search(child, name + "/" + child.name);
			    if(result != null)
			    {
				    return result;
			    }
		    }
		    return null;
	    };

	    //bones必须有数值
	    return Search(bones[rootBoneIndex], bones[rootBoneIndex].name);
    }
    
    private void NormalizeQuaternion (ref Quaternion q)
    {
	    float sum = 0;
	    for (int i = 0; i < 4; ++i)
		    sum += q[i] * q[i];
	    float magnitudeInverse = 1 / Mathf.Sqrt(sum);
	    for (int i = 0; i < 4; ++i)
		    q[i] *= magnitudeInverse;  
    }
    
    private string GetBoneHierarchyName(GPUSkinning_Bone bone)
    {
	    string str = string.Empty;

	    GPUSkinning_Bone currentBone = bone;
	    while(currentBone != null)
	    {
		    if(str == string.Empty)
		    {
			    str = currentBone.name;
		    }
		    else
		    {
			    str = currentBone.name + "/" + str;
		    }

		    currentBone = currentBone.parent;
	    }

	    return str;
    }
    
    private GPUSkinning_Bone GetBoneByTransform(Transform transform)
    {
        foreach(GPUSkinning_Bone bone in bones)
        {
            if(bone.transform == transform)
            {
                return bone;
            }
        }
        return null;
    }
}
